import numpy as np


class SudokuCore:
    def __init__(self):
        """
        :param
        """
        self.total_size = None  # size of sudoku = total_size * total_size, can be 4,9,16...
        self.grid = None
        self.cell_size = None
        self.possible_numbers = None

    def __str__(self):
        return str(self.grid)

    def init(self, size):
        """
        Init sudoku object based on size parameter
        :param size: int, defines size of sudoku grid, i.e. 4 for numbers 1-4, 9 for numbers 1-9 etc.
        :return:
        """
        self.total_size = size  # size of sudoku = total_size * total_size, can be 4,9,16...
        self.grid = np.zeros((self.total_size, ) * 2, dtype=int)
        self.cell_size = int(self.total_size ** 0.5)
        self.init_possible_numbers()

    def from_grid(self, grid):
        """
        Alternative initialization, which uses supplied sudoku grid.
        :param grid:
        :return:
        """
        self.grid = np.asarray(grid)
        self.total_size = len(grid[0])
        self.cell_size = int(self.total_size ** 0.5)
        self.reset_possible_numbers()

    def init_possible_numbers(self):
        self.possible_numbers = np.ndarray((self.total_size, ) * 3, dtype=bool)
        self.possible_numbers.fill(True)

    def get_next_free_position(self):
        """
        Return next grid position with no number assigned.
        :return:
        """
        for position in np.ndindex(self.grid.shape):
            if self.grid[position] == 0:
                return position
        return -1

    def any_free_position(self):
        return np.any(self.grid == 0)

    def get_empty_positions(self):
        return self.grid == 0

    def get_size(self):
        return self.total_size

    def get_position(self, position):
        """
        Get value of position in sudoku grid.
        :param position:
        :return:
        """
        return self.grid[position]

    def clean_position(self, position):
        self.grid[position] = 0
        self.reset_possible_numbers()

    def fill_position(self, position, value):
        self.grid[position] = value
        self.update_options_single_number(position, value)

    def reset_possible_numbers(self):
        """
        Re-generate possible numbers based on grid contents
        :return:
        """
        self.init_possible_numbers()
        for position in np.ndindex(self.grid.shape):
            if self.grid[position] != 0:
                self.update_options_single_number(position, self.grid[position])

    def update_options_single_number(self, position, number):
        """
        Update possible numbers for positions based on sudoku rules.
        :param position: tuple(x, y)
        :param number: int
        :return:
        """
        self.update_row(position[0], number)
        self.update_column(position[1], number)
        self.update_cell(position, number)

    def update_row(self, row, number):
        """
        Update possible numbers based on sudoku rule which allows only one number of a kind in a row.
        :param row:
        :param number:
        :return:
        """
        self.possible_numbers[row, :, number-1] = False

    def update_column(self, column, number):
        """
        Update possible numbers based on sudoku rule which allows only one number of a kind in a column.
        :param column:
        :param number:
        :return:
        """
        self.possible_numbers[:, column, number-1] = False

    def update_cell(self, position, number):
        """
        Update possible numbers based on sudoku rule which allows only one number of a kind in a cell.
        :param position:
        :param number:
        :return:
        """
        cell_x = (position[0] // self.cell_size) * self.cell_size
        cell_y = (position[1] // self.cell_size) * self.cell_size
        self.possible_numbers[cell_x:cell_x + self.cell_size, cell_y:cell_y + self.cell_size, number-1] = False

    def check_next_single_option_position(self):
        """
        Scan grid for next position, which has only 1 number possible as option.
        :return: coords (x, y) or -1 if not found
        """
        for position in np.ndindex(self.possible_numbers.shape[:2]):
            if sum(self.possible_numbers[position]) == 1 and self.get_position(position) == 0:
                return position
        return -1

    def check_solvable(self):
        """
        Attempt to solve sudoku. Returns True if it is possible to solve it based on implemented solving steps.
        :return:
        """
        original_grid = self.grid.copy()
        solvable = self.solve()
        self.grid = original_grid
        self.reset_possible_numbers()
        return solvable

    def solve(self):
        self.reset_possible_numbers()
        while True:
            if (position := self.check_next_single_option_position()) != -1:
                single_option_value = np.where(self.possible_numbers[position])[0] + 1
                self.fill_position(position, single_option_value)
            elif self.check_next_single_option_position() == -1:
                return not (self.grid == 0).any()



"""
Handles generation of sudoku puzzle.
"""

import numpy as np
import sudoku.sudoku_core as sudoku_core


class SudokuGenerator:
    """
    class for handling sudoku generation.
    """
    def __init__(self, size):
        """

        :param size: int, defines size of sudoku grid, i.e. 4 for numbers 1-4, 9 for numbers 1-9 etc.
        """
        self.sudoku = sudoku_core.SudokuCore()
        self.sudoku.init(size=size)
        self.number_selection_memory = []
        self.number_selection_random = []
        self.rollbacked_filter = np.ndarray((self.sudoku.get_size(), ) * 3, dtype=bool)
        self.rollbacked_filter.fill(True)

    def get_allowed_numbers(self):
        """
        Allowed numbers are combination of possible numbers (which filters numbers by the sudoku rules) and rollbacked
        filter (which filters numbers based on rollbacks done through generation process).
        :return:
        """
        return np.multiply(self.sudoku.possible_numbers, self.rollbacked_filter)

    def generate_grid(self, seed=50):
        """
        Main method for puzzle generation.
        :param seed: int => seed for numpy random function
        :return: 2d np array with sudoku values
        """
        np.random.seed(seed)
        while self.sudoku.any_free_position():
            self.randomly_fill_next_position()
            while True:
                if self.check_no_options():
                    self.rollback_last_random()
                elif (position := self.check_next_single_option_position()) != -1:
                    self.fill_single_choice(position)
                else:
                    break

    def get_sudoku(self):
        return self.sudoku

    def check_no_options(self):
        """
        Check if there are any positions, which do not have any allowed number which could be assigned to them.
        :return:
        """
        empty_positions = self.sudoku.get_empty_positions()
        positions_without_options = np.ndarray((self.sudoku.get_size(), ) * 2, dtype=bool)
        for position in np.ndindex(self.get_allowed_numbers().shape[:2]):
            positions_without_options[position] = not any(self.get_allowed_numbers()[position])
        empty_without_options = np.logical_and(empty_positions, positions_without_options)
        return np.any(empty_without_options)

    def rollback_last_random(self):
        """
        Loop through selection memory in backwards direction and remove all options, until one which was selected
        randomly was removed as well.
        :return:
        """
        for random_indicator in reversed(self.number_selection_random):
            last_random_position = self.number_selection_memory.pop()
            self.number_selection_random.pop()
            previous_value = self.sudoku.get_position(last_random_position)
            self.sudoku.clean_position(position=last_random_position)
            if random_indicator:
                rollback_index = list(last_random_position)
                rollback_index.insert(2, previous_value-1)
                rollback_index = tuple(rollback_index)
                self.rollbacked_filter[rollback_index] = False
                self.reset_dependent_rollbacks(last_random_position)
                break

    def reset_dependent_rollbacks(self, rollbacked_position):
        """
        When rollback is applied and there are some positions which were previously banned because they failed to
        produce solvable sudoku further in the line, reset them.
        :param max_position:
        :return:
        """
        positions = np.ndindex(self.rollbacked_filter.shape[:2])
        positions_filtered = []
        for position in reversed(list(positions)):
            if position == rollbacked_position:
                break
            positions_filtered.append(position)
        for position in positions_filtered:
            self.rollbacked_filter[position] = True

    def randomly_fill_next_position(self):
        """
        Select randomly number which is allowed and put it to sudoku grid.
        :return:
        """
        next_position = self.sudoku.get_next_free_position()
        if not next_position == -1:
            options = np.arange(1, self.sudoku.get_size()+1)
            allowed_mask = self.get_allowed_numbers()[next_position]
            allowed_options = options[allowed_mask]
            choice = np.random.choice(allowed_options)
            self.sudoku.fill_position(position=next_position, value=choice)
            self.number_selection_memory.append(next_position)
            self.number_selection_random.append(True)

    def check_next_single_option_position(self):
        """
        Scan grid for next position, which has only 1 number available as option.
        :return: coords (x, y) or -1 if not found
        """
        for position in np.ndindex(self.get_allowed_numbers().shape[:2]):
            if sum(self.get_allowed_numbers()[position]) == 1 and self.sudoku.get_position(position) == 0:
                return position
        return -1

    def fill_single_choice(self, position):
        """
        Used for positions which have only one possible number.
        :param position:
        :return:
        """
        single_option_value = np.where(self.get_allowed_numbers()[position])[0] + 1
        self.sudoku.fill_position(position, single_option_value)
        self.number_selection_memory.append(position)
        self.number_selection_random.append(False)



"""
Simple code generator which generates python module, test class and tests which copy class/method structure in given
module.
"""

import importlib

module_path = 'sudoku.sudoku_handler'
module_name = 'sudoku_handler'
class_name = 'SudokuHandler'

module = importlib.import_module(module_path)

test_file = f'../tests/test_{module_name}.py'
functions = [function for function in dir(getattr(module, class_name)) if function.find('__') == -1]

with open(test_file, 'w') as file:
	file.write(f"import pytest\n")
	file.write(f"import {module_path} as {module_name}\n\n\n")
	file.write(f"class Test{class_name}:\n\n")
	for function in functions:
		file.write(f"\tdef test_{function}(self):\n\t\tassert False\n\n")


"""
Tests for sudoku generation algorithm
"""
import pytest
import numpy as np
import sudoku.sudoku_core as sudoku_core
import sudoku.sudoku_generator as sudoku_generator


@pytest.fixture
def generator():
    return sudoku_generator.SudokuGenerator(size=4)


@pytest.fixture
def sudoku_fixture():
    sudoku_object = sudoku_core.SudokuCore()
    sudoku_object.init(4)
    return sudoku_object


class TestSudoku:

    def test_get_next_free_position(self, sudoku_fixture):
        sudoku_fixture.grid[0, 0] = 1
        assert sudoku_fixture.get_next_free_position() == (0, 1)

    def test_update_options_single_number(self, sudoku_fixture):
        sudoku_fixture.update_options_single_number((1, 1), 2)
        expected = np.ndarray((4, 4, 4), dtype=bool)
        expected.fill(True)
        expected[0:2, 0:2, 1] = False
        expected[:, 1, 1] = False
        expected[1, :, 1] = False
        assert (sudoku_fixture.possible_numbers == expected).all()

    def test_update_row(self, sudoku_fixture):
        sudoku_fixture.update_row(1, 2)
        expected = np.ndarray((4, 4, 4), dtype=bool)
        expected.fill(True)
        expected[1, :, 1] = False
        assert (sudoku_fixture.possible_numbers == expected).all()

    def test_update_column(self, sudoku_fixture):
        sudoku_fixture.update_column(1, 2)
        expected = np.ndarray((4, 4, 4), dtype=bool)
        expected.fill(True)
        expected[:, 1, 1] = False
        assert (sudoku_fixture.possible_numbers == expected).all()

    def test_update_cell(self, sudoku_fixture):
        sudoku_fixture.update_cell((1, 1), 2)
        expected = np.ndarray((4, 4, 4), dtype=bool)
        expected.fill(True)
        expected[0:2, 0:2, 1] = False
        assert (sudoku_fixture.possible_numbers == expected).all()

    def test_reset_possible_numbers(self, sudoku_fixture):
        sudoku_fixture.grid[1, 1] = 2
        sudoku_fixture.grid[0, 0] = 3
        sudoku_fixture.number_selection_memory = [(0, 0), (1, 1)]
        sudoku_fixture.reset_possible_numbers()
        expected = np.ndarray((4, 4, 4), dtype=bool)
        expected.fill(True)
        expected[0:2, 0:2, 1] = False
        expected[:, 1, 1] = False
        expected[1, :, 1] = False
        expected[0:2, 0:2, 2] = False
        expected[:, 0, 2] = False
        expected[0, :, 2] = False
        assert (sudoku_fixture.possible_numbers == expected).all()

    def test_check_solvable__true(self, generator):
        generator.generate_grid(50)
        generator.sudoku.grid[0, 1] = 0
        assert generator.sudoku.check_solvable()

    def test_check_solvable__false(self, sudoku_fixture):
        sudoku_fixture.grid = np.zeros((4, 4))
        sudoku_fixture.possible_numbers.fill(True)
        assert not sudoku_fixture.check_solvable()

    def test_solve(self, sudoku_fixture):
        sudoku_fixture.grid = np.asarray([[1, 2, 0, 3], [0, 0, 0, 0], [0, 0, 0, 0], [3, 0, 0, 2]])
        sudoku_fixture.reset_possible_numbers()
        sudoku_fixture.solve()
        expected = np.asarray([[1, 2, 4, 3], [4, 3, 2, 1], [2, 1, 3, 4], [3, 4, 1, 2]])
        assert (sudoku_fixture.grid == expected).all()

    def test_from_grid(self):
        sudoku_object = sudoku_core.SudokuCore()
        sudoku_object.from_grid([[1, 2, 0, 3], [0, 0, 0, 0], [0, 0, 0, 0], [3, 0, 0, 2]])
        assert sudoku_object.total_size == 4



"""
Handles preparation of sudoku for solving.
"""
import numpy as np
from enum import IntEnum


class Difficulty(IntEnum):
    easy = 10
    medium = 20
    hard = 30


class SudokuPreparer:

    def __init__(self, sudoku):
        self.sudoku = None
        self.available_for_removal = None
        self.removed_all_numbers = False
        self.sudoku = sudoku
        self.available_for_removal = np.ndarray((self.sudoku.get_size(), ) * 2, dtype=bool)
        self.available_for_removal.fill(True)

    def prepare(self, difficulty):
        """
        Prepare solved sudoku for solving by removing amount of numbers from grid based on difficulty.
        :param difficulty:
        :return:
        """
        self.remove_numbers(difficulty)
        return self.sudoku

    def remove_numbers(self, difficulty: Difficulty):
        """
        Remove numbers in sudoku grid, until either required number of removals is reached, or no more numbers can be
        removed without making sudoku unsolvable.
        :param difficulty:
        :return:
        """
        for x in range(difficulty):
            removed = self.remove_random_number()
            if not removed:
                self.removed_all_numbers = True
                break

    def remove_random_number(self):
        """
        Select random position in sudoku and remove number from it, as long as this allows sudoku to be still solvable.
        If not, rollback removal and attempt to remove other number randomly. Return true if number was successfully
        removed. In case that all the options were exhausted and still no number was removed, return False.
        case that all options are
        exhausted,
        :return:
        """
        while True:
            random_position = self.select_random_filled_position()
            previous_value = self.sudoku.grid[random_position]
            self.sudoku.grid[random_position] = 0
            solvable = self.sudoku.check_solvable()
            self.available_for_removal[random_position] = False
            if solvable:
                return True
            elif self.available_for_removal.any():
                self.sudoku.grid[random_position] = previous_value
                continue
            else:
                self.sudoku.grid[random_position] = previous_value
                return False

    def select_random_filled_position(self):
        nonzero_indices = np.nonzero(self.sudoku.grid)
        nonzero_positions = np.array(list(zip(nonzero_indices[0], nonzero_indices[1])))
        random_index = np.random.choice(len(nonzero_positions))
        random_position = tuple(nonzero_positions[random_index])
        return random_position




"""
Tests for sudoku generation algorithm
"""
import pytest
import numpy as np
import sudoku.sudoku_generator as sudoku_generator


@pytest.fixture
def generator():
    return sudoku_generator.SudokuGenerator(size=4)


class TestSudokuGenerator:

    def test_get_allowed_numbers(self, generator):
        generator.sudoku.possible_numbers[0, 0, 0] = False
        generator.rollbacked_filter[1, 1, 1] = False
        expected = np.ndarray((generator.sudoku.get_size(), generator.sudoku.get_size(), generator.sudoku.get_size()),
                              dtype=bool)
        expected.fill(True)
        expected[0, 0, 0] = expected[1, 1, 1] = False
        assert (generator.get_allowed_numbers() == expected).all()

    def test_check_no_options__false(self, generator):
        assert not generator.check_no_options()

    def test_check_no_options__true(self, generator):
        generator.sudoku.possible_numbers[2, 2, :] = False
        assert generator.check_no_options()

    def test_generate_grid(self, generator):
        generator.generate_grid()
        expected = np.asarray([[1, 2, 4, 3],
                               [4, 3, 2, 1],
                               [2, 1, 3, 4],
                               [3, 4, 1, 2]])
        assert (generator.sudoku.grid == expected).all()

    def test_randomly_fill_next_position(self, generator):
        np.random.seed(50)
        generator.randomly_fill_next_position()
        assert generator.sudoku.grid[0, 0] == 1

    def test_rollback_last_random(self, generator):
        generator.number_selection_memory = [(1, 1), (2, 2), (3, 3)]
        generator.number_selection_random = [True, True, False]
        generator.rollback_last_random()
        assert generator.number_selection_memory == [(1, 1)] and generator.number_selection_random == [True]

    def test_reset_dependent_rollbacks(self, generator):
        generator.generate_grid()
        np.put(generator.rollbacked_filter, [(1, 3), (2, 0), (2, 1)], [False, False, False, False])
        expected = generator.rollbacked_filter.copy()
        generator.reset_dependent_rollbacks((2, 0))
        expected[2, 1] = [True, True, True, True]
        assert (generator.rollbacked_filter == expected).all()

    def test_fill_single_choice(self, generator):
        generator.sudoku.possible_numbers[3, 3, :3] = False
        generator.fill_single_choice((3, 3))
        assert generator.sudoku.grid[3, 3] == 4

    def test_check_next_single_option_position__exists(self, generator):
        generator.sudoku.possible_numbers[3, 3, 1:] = False
        assert generator.check_next_single_option_position() == (3, 3)

    def test_check_next_single_option_position__none(self, generator):
        assert generator.check_next_single_option_position() == -1



import pytest
import numpy as np
import sudoku.sudoku_handler as sudoku_handler
import sudoku.sudoku_generator as sudoku_generator
import sudoku.sudoku_prepare as sudoku_prepare

@pytest.fixture
def handler():
	return sudoku_handler.SudokuHandler()


class TestSudokuHandler:

	def test_check_user_position(self, handler):
		handler.generate(size=4)
		handler.prepare_for_solving(difficulty=15)
		handler.user_grid[0, 2] = 4
		match = handler.check_user_position((0, 2))
		mismatch = handler.check_user_position((1, 0))
		assert match and not mismatch

	def test_generate(self, handler):
		handler.generate(size=4)
		expected = np.asarray([[1, 2, 4, 3],
							   [4, 3, 2, 1],
							   [2, 1, 3, 4],
							   [3, 4, 1, 2]])
		assert (expected == handler.completed_grid).all()

	def test_prepare_for_solving(self, handler):
		handler.generate(size=4)
		handler.prepare_for_solving(difficulty=15)
		expected = [[1, 2, 0, 3], [0, 0, 0, 0], [0, 0, 0, 0], [3, 0, 0, 2]]
		assert (expected == handler.sudoku.grid).all()

	def test_put_user_number(self, handler):
		handler.generate(size=4)
		handler.prepare_for_solving(difficulty=5)
		handler.put_user_number(position=[0, 1], number=5)
		assert handler.user_grid[0, 1] == 5

	def test_reset(self, handler):
		handler.generate(size=4)
		handler.prepare_for_solving(difficulty=10)
		handler.user_grid[0, 1] = 999
		handler.reset()
		assert (handler.user_grid == handler.sudoku.grid).all()

	def test_solve_grid(self, handler):
		generator = sudoku_generator.SudokuGenerator(size=4)
		generator.generate_grid()
		grid = generator.sudoku.grid
		preparer = sudoku_prepare.SudokuPreparer(sudoku=generator.sudoku)
		preparer.prepare(difficulty=15)
		handler.solve_grid(preparer.sudoku.grid)
		assert (handler.completed_grid == grid).all()



"""
Sudoku API
"""

from sudoku.sudoku_core import SudokuCore
from sudoku.sudoku_prepare import SudokuPreparer
from sudoku.sudoku_generator import SudokuGenerator


class SudokuHandler:

	def __init__(self):
		self.completed_grid = None
		self.sudoku = None
		self.user_grid = None

	def generate(self, size):
		"""
		generate sudoku object
		:return:
		"""
		generator = SudokuGenerator(size=size)
		generator.generate_grid()
		self.sudoku = generator.sudoku
		self.completed_grid = generator.sudoku.grid

	def prepare_for_solving(self, difficulty):
		"""
		prepare sudoku so it can be solved
		:return:
		"""
		preparer = SudokuPreparer(sudoku=self.sudoku)
		self.sudoku = preparer.prepare(difficulty=difficulty)
		self.user_grid = self.sudoku.grid

	def get_grid(self):
		"""
		return sudoku grid
		:return:
		"""
		return self.completed_grid

	def reset(self):
		"""
		resets sudoku to original state before user stared solving
		:return:
		"""
		self.user_grid = self.sudoku.grid

	def put_user_number(self, position, number):
		"""
		puts number by the user into grid
		:return:
		"""
		self.user_grid[position] = number
		pass

	def check_user_position(self, position):
		"""
		checks if number put by user is correct
		:return:
		"""
		return self.completed_grid[position] == self.user_grid[position]

	def solve_grid(self, grid):
		"""
		Creates sudoku object from supplied grid and solves it if possible.
		:return:
		"""
		sudoku = SudokuCore()
		sudoku.from_grid(grid)
		sudoku.solve()
		self.completed_grid = sudoku.grid



from setuptools import setup

setup(
    name='Sudoku-py-lib',
    version='1.0.1',
    packages=['tests', 'sudoku'],
    url='',
    license='Apache 2.0',
    author='Vollinger',
    author_email='',
    description='package to generate and solve sudoku puzzles'
)



import numpy as np
import pytest
import sudoku.sudoku_prepare as sudoku_prepare
import sudoku.sudoku_generator as sudoku_generator


@pytest.fixture
def sudoku_preparer():
    generator = sudoku_generator.SudokuGenerator(size=4)
    generator.generate_grid()
    prepared = sudoku_prepare.SudokuPreparer(sudoku=generator.sudoku)
    return prepared


class TestSudokuPreparer:

    def test_prepare(self, sudoku_preparer):
        expected = [[1, 2, 0, 3], [0, 0, 0, 0], [0, 0, 0, 0], [3, 0, 0, 2]]
        sudoku_preparer.prepare(15)
        assert (sudoku_preparer.sudoku.grid == expected).all()

    def test_remove_random_number(self,  sudoku_preparer):
        assert sudoku_preparer.remove_random_number()
